from collections import deque
import typing
import warnings

from ponty.memo.cache.base import CacheStore, cache, cachemiss, Stampede
from ponty.memo.lock.local import locallock
from ponty.utils import now_millis


T = typing.TypeVar("T")


class _LRUStore(CacheStore[T]):

    def __init__(self, ttl_ms: int, maxsize: int):
        super().__init__()

        if ttl_ms <= 0:
            raise ValueError("ttl must be positive")
        self._ttl_ms: typing.Final[int] = ttl_ms

        if maxsize < 0:
            raise ValueError("negative maxsize is nonsense")
        if maxsize == 0:
            warnings.warn("Cache without maxsize may grow endlessly, beware.")
        self._maxsize: typing.Final[int] = maxsize

        self._cache: dict[str, tuple[int, T]] = dict()
        self._lru: deque[str] = deque()

    async def get(self, key: str) -> typing.Union[T, type[cachemiss]]:
        try:
            expiry, data = self._cache[key]
        except KeyError:
            return cachemiss

        if expiry < now_millis():
            await self.remove(key)
            return cachemiss

        # rotate key to the end of the LRU queue
        self._lru.remove(key)
        self._lru.append(key)

        return data

    async def set(self, key: str, data: T) -> None:
        expiry = now_millis() + self._ttl_ms
        self._cache[key] = (expiry, data)
        self._lru.append(key)
        await self._sizecheck()

    async def remove(self, key: str) -> bool:
        try:
            self._cache.pop(key)
        except KeyError:
            return False
        else:
            self._lru.remove(key)
        return True

    async def _sizecheck(self) -> None:
        if not self._maxsize:
            return

        while len(self._lru) > self._maxsize:
            key = self._lru.popleft()
            self._cache.pop(key)


def localcache(
    *,
    ttl_ms: int,
    maxwait_ms: int = 1000,
    pulse_ms: int = 50,
    maxsize: int = 20,
    name: str = "",
):
    """Process-RAM LRU memoizer with antistampede.

    Good for small frequently-used datasets (high ttl) or volatile
    stampede-likely objects (low ttl).

    ttl_ms: millis to expiry. Must be positive
    maxwait_ms: millis to wait for a lock to resolve.
                Throws "Stampede" error when (n * pulse) > maxwait
    pulse_ms: antistampede recheck frequency
    maxsize: evict least recently used values once the cache reaches <maxsize> elements
    name: providing a name registers the cache, so it can be used by 'invalidate'

    """
    antistampede = locallock(
        maxwait_ms=maxwait_ms,
        pulse_ms=pulse_ms,
        timeout_error=Stampede,
    )

    return cache(
        store=_LRUStore(ttl_ms, maxsize),
        antistampede=antistampede,
        name=name,
    )



import abc
from contextlib import asynccontextmanager
import functools
import hashlib
import json
import typing
import warnings

from ponty.memo.lock.base import Lock, Locked
from ponty.registry import Registry


class Stampede(Locked): ...


class cachemiss: ...




T = typing.TypeVar("T")


class CacheStore(abc.ABC, typing.Generic[T]):

    @abc.abstractmethod
    async def get(self, key: str) -> typing.Union[T, type[cachemiss]]:
        """Return the cached value or the `cachemiss` sentinel. Errors are not captured."""

    @abc.abstractmethod
    async def set(self, key: str, data: T) -> None: ...

    @abc.abstractmethod
    async def remove(self, key: str) -> bool: ...




_registry = Registry[tuple[CacheStore, Lock]]()


def cache(store: CacheStore, antistampede: Lock, name: str = ""):
    """Cache the decorated function's return value. Not thread-safe.

    store: container (or proxy), implementing get/set mechanics for cached items
    antistampede: guard class, blocks on simultaneous access to the requested
                  key to avoid a stampede
    name: providing a name registers the cache, so it can be used by 'invalidate' below

    """
    def wraps(f):
        if name:
            _registry.add(name, (store, antistampede))

        @functools.wraps(f)
        async def wrapper(*a, **kw):
            key = _hash_args(*a, **kw)

            async with antistampede.lock(key):
                data = await store.get(key)
                if data is not cachemiss:
                    return data

                data = await f(*a, **kw)
                await store.set(key, data)

            return data

        return wrapper
    return wraps


@asynccontextmanager
async def invalidate(cachename: str, *a, **kw) -> typing.AsyncIterator[None]:
    """Invalidate an item in cache <cachename>.

    Waits for and holds the cache's lock until the context manager exits.
    This prevents simultaneous fetches from re-cache-ing the item before
    mutations are committed. This also means functions calling 'invalidate'
    CANNOT be wrapped with a cache decorator, as it will already hold the lock.

    Note the *arg/**kwarg combo must match the arguments to the cache-decorated
    function EXACTLY in order to get a key hit. E.g.

    @localcache(name="foo", ttl_ms=60000)
    async def get_foo(foo_id):
        ...

    async def update_foo(foo_id, ...):
        async with invalidate("foo", foo_id):  # this hits
            ...
        async with invalidate("foo", foo_id=foo_id):  # this does not
            ...

    """
    key = _hash_args(*a, **kw)

    try:
        store, antistampede = _registry.get(cachename)
    except KeyError:
        warnings.warn(f"'{cachename}' is not registered", category=RuntimeWarning)
        yield
    else:
        async with antistampede.lock(key):
            await store.remove(key)
            yield


def _hash_args(*a, **kw) -> str:
    args = a + tuple(sorted(kw.items()))
    encoded = json.dumps(args).encode("utf-8")
    return hashlib.sha1(encoded).hexdigest()



import functools
import inspect

import aiohttp.web

from ponty.http.expect.body import (
    ValidatedJsonBody,
    ParsedJsonBody,
)
from ponty.http.expect.header import (
    Header,
    ContentLength,
    ContentType,
)
from ponty.http.expect.query import QueryParameter, QueryParameterEnum
from ponty.http.expect.req import (
    expect,
    Request,
    JsonBody,
    TextBody,
)
from ponty.http.expect.route import (
    RouteParameter,
    PosIntRouteParameter,
    StringRouteParameter,
)


class AIOHttpReq:

    def __get__(self, obj, objtype=None) -> aiohttp.web.Request:
        return obj.req


class Cookie:

    def __init__(self, *, name: str):
        self._name = name

    def __get__(self, obj, objtype=None) -> str:
        return obj.req.cookies[self._name]



import asyncio
from contextlib import asynccontextmanager
import typing

import aiohttp

from ponty.http.go import Provider
from ponty.registry import Registry


_registry = Registry[tuple[asyncio.Semaphore, aiohttp.ClientSession]]()


def http_client_provider(
    *,
    name: str,
    timeout: int = 5,
    concurrency: int = 100,
    headers: dict[str, typing.Any] = None,
    cookies: dict[str, typing.Any] = None,
    **kw
) -> Provider:

    async def provider(_) -> typing.AsyncIterator[None]:
        sem = asyncio.Semaphore(concurrency)
        sess = aiohttp.ClientSession(
            timeout=aiohttp.ClientTimeout(total=timeout),
            headers=headers,
            cookies=cookies,
            **kw
        )

        _registry.add(name, (sem, sess))

        try:
            yield
        finally:
            await sess.close()
            await asyncio.sleep(0.5)  # sleep a beat to allow underlying connections to close

    return provider


@asynccontextmanager
async def lease_http_client(name: str) -> typing.AsyncIterator[aiohttp.ClientSession]:
    sem, sess = _registry.get(name)
    async with sem:
        yield sess



import functools

import aiohttp.web

from ponty.utils import now_millis


def render_json(f):
    @functools.wraps(f)
    async def wrapper(*a, **kw) -> aiohttp.web.Response:
        return aiohttp.web.json_response({
            "now": (start := now_millis()),
            "data": await f(*a, **kw),
            "elapsed": now_millis() - start,
        })
    return wrapper



import typing

from ponty.http.expect.req import Request


T = typing.TypeVar("T")


class RouteParameter(typing.Generic[T]):

    def __init__(
        self,
        *,
        pattern: str,
        cast_to: typing.Callable[[str], T] = None,
    ):
        self._pattern = pattern
        self._cast_func = cast_to

    def __set_name__(self, owner: type[Request], name: str):
        self._key = name

    @typing.overload
    def __get__(self, obj: None, objtype: type[Request]) -> str: ...

    @typing.overload
    def __get__(self, obj: Request, objtype: type[Request]) -> T: ...

    def __get__(
        self,
        obj: typing.Optional[Request],
        objtype: type[Request] = None,
    ) -> typing.Union[T, str]:

        if obj is None:
            return f"{{{self._key}:{self._pattern}}}"

        val = obj.req.match_info[self._key]
        if self._cast_func:
            return self._cast_func(val)
        return val


class PosIntRouteParameter(RouteParameter[int]):

    def __init__(self):
        super().__init__(pattern="\d+", cast_to=int)


class StringRouteParameter(RouteParameter[str]):

    def __init__(self):
        super().__init__(pattern="\w+")



import json
import typing

import aiohttp.web
import jsonschema  # type: ignore

from ponty.http.schema import (
    dataclass_to_jsonschema,
    validator_for_schema,
    validator_from_file,
)
from ponty.http.expect.req import JsonBody, Request
from ponty.types_ import DataclassProtocol
from ponty.utils import logical_xor


class ValidatedJsonBody(JsonBody):

    def __init__(
        self,
        *,
        filepath: str = None,
        schema: dict[str, typing.Any] = None,
    ):
        super().__init__()

        if not logical_xor(filepath, schema):
            raise RuntimeError("only one of filepath, schema may be specified")

        if filepath:
            self._validator = validator_from_file(filepath)
        else:
            self._validator = validator_for_schema(schema)

    def __get__(self, obj: Request, objtype: type[Request]):
        body = super().__get__(obj, objtype)

        try:
            self._validator.validate(body)
        except jsonschema.ValidationError as e:
            raise aiohttp.web.HTTPBadRequest(text=e.message)

        return body


D = typing.TypeVar("D", bound=DataclassProtocol)


class ParsedJsonBody(ValidatedJsonBody, typing.Generic[D]):

    def __init__(self, cls: type[D], filepath: str = None):
        self._cls = cls

        if filepath:
            super().__init__(filepath=filepath)
        else:
            super().__init__(schema=dataclass_to_jsonschema(cls))

    def __get__(self, obj: Request, objtype: type[Request]) -> D:
        body = super().__get__(obj, objtype)
        return self._cls(**body)



from ponty.memo.lock.base import SentinelStore, Lock


class _Sentinels(SentinelStore):

    def __init__(self):
        super().__init__()
        self._sentinels: set[str] = set()

    async def exists(self, key: str) -> bool:
        return key in self._sentinels

    async def add(self, key: str) -> None:
        self._sentinels.add(key)

    async def remove(self, key: str) -> None:
        self._sentinels.remove(key)


def locallock(**kw) -> Lock:
    """Mutex. Coordinates across coroutines. Not thread-safe."""
    return Lock(sentinels=_Sentinels(), **kw)



__all__ = ["dataclass_to_jsonschema", "Annotation"]


import dataclasses
from dataclasses import MISSING
import datetime
import enum
import numbers
import typing

from ponty.types_ import DataclassProtocol


class Annotation(dict[str, typing.Any]): ...




D = typing.TypeVar("D", bound=DataclassProtocol)


def dataclass_to_jsonschema(cls: type[D]) -> dict[str, typing.Any]:
    """Generate the jsonschema for a dataclass. Supports backreferences.

    Optionally, fields can be enriched by passing jsonschema keywords to Annotation.

    Additionally, the dataclass can be enriched with keywords by embedding
    a JsonSchema class (using Annotation as well).

    E.g.,


    @dataclasses.dataclass
    class Model:

        name: str
        body_style: typing.Literal["coupe", "sedan", "hatchback", "convertible", "wagon", "suv"]
        year: typing.Annotated[int, Annotation(minimum=1900, maximum=2050)]
        msrp: typing.Annotated[int, Annotation(description="MSRP, dollars.")]
        cylinders: int = 4


    @dataclasses.dataclass
    class Make:

        name: typing.Annotated[str, Annotation(description="Brand name")]
        models: list[Model]

        class JsonSchema:
            annotation = Annotation(description="Brand name", examples=["Toyota", "Honda"])

    """
    if not dataclasses.is_dataclass(cls):
        raise ValueError("Dataclasses only.")

    defs: dict[str, typing.Any] = {}
    schema = _cache_dataclass(cls, cls, defs, True)

    if defs:
        schema["$defs"] = defs

    return {
        "$schema": "http://json-schema.org/draft-07/schema#",
        **schema,
    }




def _cache_dataclass(cls, primary, defs, firstpass=False):
    if cls is primary:
        if firstpass:
            return _render_dataclass(cls, primary, defs)
        return {"allOf": [{"$ref": "#"}]}

    name = cls.__name__
    if name not in defs:
        defs[name] = _render_dataclass(cls, primary, defs)

    return {"allOf": [{"$ref": f"#/$defs/{name}"}]}


def _render_dataclass(cls, primary, defs):
    properties = {}
    required = []

    for field in dataclasses.fields(cls):
        schema = _lookup(field.type, primary, defs)
        if field.default is not MISSING:
            schema["default"] = field.default

        properties[field.name] = schema

        if field.default is MISSING and field.default_factory is MISSING:
            required.append(field.name)

    schema = {
        "properties": properties,
        "title": cls.__name__,
        "type": "object",
        **getattr(getattr(cls, "JsonSchema", {}), "annotation", {})
    }

    if required:
        schema["required"] = required

    return schema


def _lookup(field_type, primary, defs, **kw):
    if dataclasses.is_dataclass(field_type):
        return _cache_dataclass(field_type, primary, defs)

    if field_type is typing.Any:
        return {}

    if field_type is None or field_type is type(None):
        return {"type": "null", **kw}

    unscripted_type = typing.get_origin(field_type)
    args = typing.get_args(field_type)

    if unscripted_type is typing.Annotated:
        for anno in args[1:]:
            if isinstance(anno, Annotation):
                break
        else:
            anno = {}

        return _lookup(args[0], primary, defs, **anno)

    if unscripted_type is typing.Literal:
        return {"enum": list(args), **kw}

    if unscripted_type is typing.Union:  # also catches Optional
        # TODO list would be nicer? this is prob more reliable
        return {
            "anyOf": [_lookup(arg, primary, defs) for arg in args],
            **kw
        }

    if field_type is bool:
        return {"type": "boolean", **kw}
    if field_type is int:
        return {"type": "integer", **kw}
    if field_type is str:
        return {"type": "string", **kw}

    if field_type is dict or unscripted_type is dict:
        schema = {"type": "object", **kw}
        if len(args) == 2:
            t1, t2 = args
            if t1 is not str:
                raise Exception
            schema["additionalProperties"] = _lookup(t2, primary, defs)
        return schema

    if field_type is list or unscripted_type is list:
        schema = {"type": "array", **kw}
        if len(args) == 1:
            schema["items"] = _lookup(args[0], primary, defs)
        return schema

    if field_type is set or unscripted_type is set:
        schema = {"type": "array", "uniqueItems": True, **kw}
        if args:
            schema["items"] = _lookup(args[0], primary, defs)
        return schema

    if field_type is tuple or unscripted_type is tuple:
        schema = {"type": "array", **kw}

        if args:
            if len(args) == 2 and (args[1] is ...):
                schema["items"] = _lookup(args[0], primary, defs)
            else:
                schema["prefixItems"] = [_lookup(arg, primary, defs) for arg in args]
                schema["minItems"] = schema["maxItems"] = len(args)

        return schema

    if issubclass(field_type, datetime.datetime):  # subclass of datetime.date, so must come first
        return {"type": "string", "format": "date-time", **kw}
    if issubclass(field_type, datetime.date):
        return {"type": "string", "format": "date", **kw}
    if issubclass(field_type, numbers.Number):
        return {"type": "number", **kw}

    if issubclass(field_type, enum.Enum):
        name = field_type.__name__
        if name not in defs:
            defs[name] = {
                "title": name,
                "enum": [i.value for i in field_type],
                **kw
            }
        return {"allOf": [{"$ref": f"#/$defs/{name}"}]}

    raise Exception(f"no support for {field_type}/{unscripted_type}")


if __name__ == "__main__":
    import json

    @dataclasses.dataclass
    class Bar:
        x: int
        y: int

    @dataclasses.dataclass
    class Foo:
        name: typing.Annotated[str, Annotation(minLength=22, pattern=".*")]
        age: int
        blah: list[int]
        blargh: dict[str, Bar]
        blam: tuple[Bar, ...]
        favorite_color: typing.Optional[str]
        union_test: typing.Union[str, int] = 7

        class JsonSchema:
            annotation = Annotation(description="go you!")

    schema = dataclass_to_jsonschema(Foo)
    print(json.dumps(schema, indent=2))



import typing

import aiohttp.web


class QueryParameter:

    def __init__(
        self,
        *,
        key: str = None,
        required: bool = False,
        default: str = "",
    ):
        self._key = key
        self._required = required
        self._default = default

    def __set_name__(self, obj, name):
        if not self._key:
            self._key = name

    def __get__(self, obj, objtype=None) -> str:
        try:
            return obj.req.query[self._key]
        except KeyError:
            if self._required:
                raise aiohttp.web.HTTPBadRequest
            return self._default


class QueryParameterEnum(QueryParameter):

    def __init__(self, *, values: typing.Iterable[str], **kw):
        super().__init__(**kw)
        self._values = values

        if self._default and self._default not in values:
            raise ValueError(f"default '{self._default}' does not validate")

    def __get__(self, obj, objtype=None) -> str:
        val = super().__get__(obj, objtype)
        if val not in self._values:
            msg = f"{self._key} must be one of {{{','.join(self._values)}}}"
            raise aiohttp.web.HTTPBadRequest(text=msg)
        return val



import abc
import asyncio
from contextlib import asynccontextmanager
from dataclasses import dataclass
import typing

from ponty.errors import Locked


class SentinelStore(abc.ABC):

    @abc.abstractmethod
    async def exists(self, key: str) -> bool: ...

    @abc.abstractmethod
    async def add(self, key: str) -> None: ...

    @abc.abstractmethod
    async def remove(self, key: str) -> None: ...


@dataclass(frozen=True)
class Lock:
    """Basic mechanic for locking / antistampede. Not thread-safe.

    Operates as a mandatory lock by default. Control blocking by specifying
    the <maxwait_ms> parameter.

    sentinels: locker, providing add/remove operations and existence checking
    maxwait_ms: patience, in millis. Raises <timeout_error> when n * pulse > maxwait
    pulse_ms: attempt to acquire the lock approx every <pulse> milliseconds
    timeout_error: exception class raised when <maxwait_ms> exceeded.
                   If provided, must be a subclass of Locked.

    """
    sentinels: SentinelStore
    maxwait_ms: int = 0
    pulse_ms: int = 100
    timeout_error: type[Locked] = Locked

    @asynccontextmanager
    async def lock(self, key: str) -> typing.AsyncIterator[None]:
        elapsed = 0
        is_locked = await self.sentinels.exists(key)

        while is_locked and elapsed < self.maxwait_ms:
            await asyncio.sleep(self.pulse_ms * 0.001)
            elapsed += self.pulse_ms
            is_locked = await self.sentinels.exists(key)

        if is_locked:
            raise self.timeout_error(text=key)

        try:
            await self.sentinels.add(key)
            yield
        finally:
            await self.sentinels.remove(key)
